"""Default GitHub event dispatcher."""

from functools import wraps

from .routers import ConcurrentRouter


__all__ = (
    'dispatch_event',
    'process_event',
    'process_event_actions',
    'WEBHOOK_EVENTS_ROUTER',
)


WEBHOOK_EVENTS_ROUTER = ConcurrentRouter()
"""An event dispatcher for webhooks."""


dispatch_event = WEBHOOK_EVENTS_ROUTER.dispatch  # pylint: disable=invalid-name
process_event = WEBHOOK_EVENTS_ROUTER.register  # pylint: disable=invalid-name


def process_event_actions(event_name, actions=None):
    """Subscribe to multiple events."""
    if actions is None:
        actions = []

    def decorator(original_function):

        def wrapper(*args, **kwargs):
            return original_function(*args, **kwargs)

        if not actions:
            wrapper = process_event(event_name)(wrapper)

        for action in actions:
            wrapper = process_event(event_name, action=action)(wrapper)

        return wraps(original_function)(wrapper)

    return decorator



"""Utility helpers for CLI."""

import contextlib
import itertools
import json
from uuid import UUID, uuid4

import multidict
import yaml


def _probe_yaml(event_file_fd):
    try:
        http_headers, event, extra = itertools.islice(
            itertools.chain(
                yaml.safe_load_all(event_file_fd),
                (None, ) * 3,
            ),
            3,
        )
    except yaml.parser.ParserError as yaml_err:
        raise ValueError('YAML file is not valid') from yaml_err
    finally:
        event_file_fd.seek(0)

    if extra is not None:
        raise ValueError('YAML file must only contain 1–2 documents')

    if event is None:
        event = http_headers
        http_headers = ()

    if event is None:
        raise ValueError('YAML file must contain 1–2 non-empty documents')

    return http_headers, event


def _probe_jsonl(event_file_fd):
    event = None

    first_line = event_file_fd.readline()
    second_line = event_file_fd.readline()
    third_line = event_file_fd.readline()
    event_file_fd.seek(0)

    if third_line:
        raise ValueError('JSONL file must only contain 1–2 JSON lines')

    http_headers = json.loads(first_line)

    with contextlib.suppress(ValueError):
        event = json.loads(second_line)

    if event is None:
        event = http_headers
        http_headers = ()

    return http_headers, event


def _probe_json(event_file_fd):
    event = json.load(event_file_fd)
    event_file_fd.seek(0)

    if not isinstance(event, dict):
        raise ValueError('JSON file must only contain an object mapping')

    http_headers = ()

    return http_headers, event


def _parse_fd_content(event_file_fd):
    """Guess file content type and read event with HTTP headers."""
    for event_reader in _probe_yaml, _probe_jsonl, _probe_json:
        with contextlib.suppress(ValueError):
            return event_reader(event_file_fd)

    raise ValueError(
        'The input event VCR file has invalid structure. '
        'It must be either of YAML, JSONL or JSON.',
    )


def _transform_http_headers_list_to_multidict(headers):
    if isinstance(headers, dict):
        raise ValueError(
            'Headers must be a sequence of mappings because keys can repeat',
        )
    return multidict.CIMultiDict(next(iter(h.items()), ()) for h in headers)


def parse_event_stub_from_fd(event_file_fd):
    """Read event with HTTP headers as CIMultiDict instance."""
    http_headers, event = _parse_fd_content(event_file_fd)
    return _transform_http_headers_list_to_multidict(http_headers), event


def validate_http_headers(headers):
    """Verify that HTTP headers look sane."""
    if headers['content-type'] != 'application/json':
        raise ValueError("Content-Type must be 'application/json'")

    if not headers['user-agent'].startswith('GitHub-Hookshot/'):
        raise ValueError("User-Agent must start with 'GitHub-Hookshot/'")

    x_gh_delivery_exc = ValueError('X-GitHub-Delivery must be of type UUID4')
    try:
        x_gh_delivery_uuid = UUID(headers['x-github-delivery'])
    except ValueError as val_err:
        raise x_gh_delivery_exc from val_err
    if x_gh_delivery_uuid.version != 4:
        raise x_gh_delivery_exc

    if not isinstance(headers['x-github-event'], str):
        raise ValueError('X-GitHub-Event must be a string')


def augment_http_headers(headers):
    """Add fake HTTP headers for the missing positions."""
    fake_headers = make_http_headers_from_event(headers['x-github-event'])

    if 'content-type' not in headers:
        headers['content-type'] = fake_headers['content-type']

    if 'user-agent' not in headers:
        headers['user-agent'] = fake_headers['user-agent']

    if 'x-github-delivery' not in headers:
        headers['x-github-delivery'] = fake_headers['x-github-delivery']

    return headers


def make_http_headers_from_event(event_name):
    """Generate fake HTTP headers with the given event name."""
    return multidict.CIMultiDict({
        'content-type': 'application/json',
        'user-agent': 'GitHub-Hookshot/fallback-value',
        'x-github-delivery': str(uuid4()),
        'x-github-event': event_name,
    })



"""Shared fixtures for tests."""
import pytest

from cryptography.hazmat.backends import default_backend
from cryptography.hazmat.primitives.asymmetric.rsa import generate_private_key
from cryptography.hazmat.primitives.serialization import (
    Encoding,
    NoEncryption,
    PrivateFormat,
)


@pytest.fixture
def rsa_private_key():
    """Generate an RSA private key."""
    return generate_private_key(
        public_exponent=65537,
        key_size=4096,
        backend=default_backend(),
    )


@pytest.fixture
def rsa_private_key_bytes(rsa_private_key) -> bytes:
    r"""Generate an unencrypted PKCS#1 formatted RSA private key.

    Encoded as PEM-bytes.

    This is what the GitHub-downloaded PEM files contain.

    Ref: https://developer.github.com/apps/building-github-apps/\
         authenticating-with-github-apps/
    """
    return rsa_private_key.private_bytes(
        encoding=Encoding.PEM,
        format=PrivateFormat.TraditionalOpenSSL,  # A.K.A. PKCS#1
        encryption_algorithm=NoEncryption(),
    )



"""Runtime context helpers."""

from __future__ import annotations

from contextvars import ContextVar, Token
import typing


class ContextLookupError(AttributeError):
    """Context var lookup error."""


class _ContextMap:
    __slots__ = '__map__', '__token_map__'

    def __init__(self, **initial_vars):
        self.__map__: typing.Dict[str, ContextVar[typing.Any]] = {
            k: ContextVar(v) for k, v in initial_vars.items()
        }
        """Storage for all context vars."""

        self.__token_map__: typing.Dict[str, Token[typing.Any]] = {}
        """Storage for individual context var reset tokens."""

    def __dir__(self):
        """Render a list of public attributes."""
        return self.__map__.keys()

    def __getattr__(self, name):
        if name in ('__map__', '__token_map__'):
            return getattr(self, name)
        try:
            return self.__map__[name].get()
        except LookupError:
            raise ContextLookupError(
                f'No `{name}` present in the context',
            ) from None

    def __setattr__(self, name, value):
        if name in ('__map__', '__token_map__'):
            object.__setattr__(self, name, value)
        elif name in self.__map__:
            reset_token = self.__map__[name].set(value)
            self.__token_map__[name] = reset_token
        else:
            raise ContextLookupError(f'No `{name}` present in the context')

    def __delattr__(self, name):
        if name not in self.__map__:
            raise ContextLookupError(f'No `{name}` present in the context')
        reset_token = self.__token_map__[name]
        self.__map__[name].reset(reset_token)
        del self.__token_map__[name]



"""Generic GitHub event containers."""

from __future__ import annotations

import json
import pathlib
from typing import Any, Mapping, TextIO, TYPE_CHECKING, Union
import uuid
import warnings

import attr
from gidgethub.sansio import Event as _GidgetHubEvent

# pylint: disable=relative-beyond-top-level
from ...utils.asynctools import aio_gather
# pylint: disable=relative-beyond-top-level
from ..utils.event_utils import (
    augment_http_headers,
    parse_event_stub_from_fd,
    validate_http_headers,
)

if TYPE_CHECKING:
    # pylint: disable=relative-beyond-top-level
    from ...app.routing.abc import OctomachineryRouterBase


__all__ = 'GitHubEvent', 'GitHubWebhookEvent'


def _to_uuid4(value: Union[str, uuid.UUID]) -> uuid.UUID:
    """Return a UUID from the value."""
    if isinstance(value, uuid.UUID):
        return value

    return uuid.UUID(value, version=4)


def _to_dict(value: Union[dict, bytes, str]) -> dict:
    """Return a dict from the value."""
    if isinstance(value, dict):
        return value

    if isinstance(value, bytes):
        value = value.decode()

    return json.loads(value)


@attr.dataclass(frozen=True)
class GitHubEvent:
    """Representation of a generic source-agnostic GitHub event."""

    name: str
    """Event name."""
    payload: dict = attr.ib(converter=_to_dict)
    """Event payload object."""

    # pylint: disable=no-self-use
    @payload.validator
    def _is_payload_dict(self, attribute: str, value: dict) -> None:
        """Verify that the attribute value is a dict.

        :raises ValueError: if it's not
        """
        if isinstance(value, dict):
            return

        raise ValueError(
            f'{value!r} is passed as {attribute!s} but it must '
            'be an instance of dict',
        )

    @classmethod
    def from_file(
            cls: GitHubEvent,
            event_name: str,
            event_path: Union[pathlib.Path, str],
    ) -> GitHubEvent:
        """Construct a GitHubEvent instance from event name and file."""
        # NOTE: This could be async but it probably doesn't matter
        # NOTE: since it's called just once during init and GitHub
        # NOTE: Action runtime only has one event to process
        # NOTE: OTOH it may slow-down tests parallelism
        # NOTE: so may deserve to be fixed
        with pathlib.Path(event_path).open() as event_source:
            return cls(event_name, json.load(event_source))

    @classmethod
    def from_fixture_fd(
            cls: GitHubEvent,
            event_fixture_fd: TextIO,
            *,
            event: str = None,
    ) -> GitHubEvent:
        """Make a GitHubEvent from a fixture fd and an optional name."""
        headers, payload = parse_event_stub_from_fd(event_fixture_fd)
        if event and 'x-github-event' in headers:
            raise ValueError(
                'Supply only one of an event name '
                'or an event header in the fixture file',
            )
        event_name = event or headers['x-github-event']
        return cls(event_name, payload)

    @classmethod
    def from_fixture(
            cls: GitHubEvent,
            event_fixture_path: Union[pathlib.Path, str],
            *,
            event: str = None,
    ) -> GitHubEvent:
        """Make a GitHubEvent from a fixture and an optional name."""
        with pathlib.Path(event_fixture_path).open() as event_source:
            return cls.from_fixture_fd(event_source, event=event)

    @classmethod
    def from_gidgethub(cls, event: _GidgetHubEvent) -> GitHubEvent:
        """Construct GitHubEvent from from GidgetHub Event."""
        return cls(
            name=event.event,
            payload=event.data,
        )

    def to_gidgethub(self) -> _GidgetHubEvent:
        """Produce GidgetHub Event from self."""
        return _GidgetHubEvent(
            data=self.payload,
            event=self.name,
            delivery_id=str(uuid.uuid4()),
        )

    # pylint: disable=no-self-use
    async def dispatch_via(
            self,
            *routers: OctomachineryRouterBase,
            ctx: Mapping[str, Any] = None,
    ) -> None:
        """Invoke this event handlers from different routers."""
        if not routers:
            raise ValueError('At least one router must be supplied')

        if ctx is None:
            ctx = {}

        await aio_gather(*(
            r.dispatch(self, **ctx)
            for r in routers
        ))


@attr.dataclass(frozen=True)
class GitHubWebhookEvent(GitHubEvent):
    """Representation of a GitHub event arriving by HTTP."""

    delivery_id: uuid.UUID = attr.ib(converter=_to_uuid4)
    """A unique UUID4 identifier of the event delivery on GH side."""

    # pylint: disable=no-self-use
    @delivery_id.validator
    def _is_delivery_id(self, attribute: str, value: uuid.UUID) -> None:
        """Verify that the attribute value is UUID v4.

        :raises ValueError: if it's not
        """
        if isinstance(value, uuid.UUID) and value.version == 4:
            return

        raise ValueError(
            f'{value!r} is passed as {attribute!s} but it must '
            'be an instance of UUID v4',
        )

    @classmethod
    def from_file(
            cls: GitHubWebhookEvent,
            event_name: str,
            event_path: Union[pathlib.Path, str],
    ) -> GitHubWebhookEvent:
        """Explode when constructing from file."""
        raise RuntimeError(
            'Webhook event is not supposed to be constructed from a file',
        )

    @classmethod
    def from_fixture_fd(
            cls: GitHubWebhookEvent,
            event_fixture_fd: TextIO,
            *,
            event: str = None,
    ) -> GitHubWebhookEvent:
        """Make GitHubWebhookEvent from fixture fd and optional name."""
        headers, payload = parse_event_stub_from_fd(event_fixture_fd)
        if event and 'x-github-event' in headers:
            raise ValueError(
                'Supply only one of an event name '
                'or an event header in the fixture file',
            )
        headers['x-github-event'] = event or headers['x-github-event']
        headers = augment_http_headers(headers)
        validate_http_headers(headers)

        return cls(
            name=headers['x-github-event'],
            payload=payload,
            delivery_id=headers['x-github-delivery'],
        )

    @classmethod
    def from_fixture(
            cls: GitHubWebhookEvent,
            event_fixture_path: Union[pathlib.Path, str],
            *,
            event: str = None,
    ) -> GitHubWebhookEvent:
        """Make a GitHubWebhookEvent from fixture and optional name."""
        with pathlib.Path(event_fixture_path).open() as event_source:
            return cls.from_fixture_fd(event_source, event=event)

    @classmethod
    def from_http_request(
            cls: GitHubWebhookEvent,
            http_req_headers: Mapping[str, str],
            http_req_body: bytes,
    ):
        """Make a GitHubWebhookEvent from HTTP req headers and body."""
        return cls(
            name=http_req_headers['x-github-event'],
            payload=json.loads(http_req_body.decode()),
            delivery_id=http_req_headers['x-github-delivery'],
        )

    @classmethod
    def from_gidgethub(cls, event: _GidgetHubEvent) -> GitHubWebhookEvent:
        """Construct GitHubWebhookEvent from from GidgetHub Event."""
        return cls(
            name=event.event,
            payload=event.data,
            delivery_id=event.delivery_id,
        )

    def to_gidgethub(self) -> _GidgetHubEvent:
        """Produce GidgetHub Event from self."""
        return _GidgetHubEvent(
            data=self.payload,
            event=self.name,  # pylint: disable=no-member
            delivery_id=self.delivery_id,
        )


class GidgetHubEventMixin:
    """A temporary shim for GidgetHub event interfaces.

    It's designed to be used during the refactoring period when interfacing
    with the new event representation layer :py:class:`~GitHubEvent`.
    """

    @property
    def data(self):
        """Event payload dict alias."""
        warnings.warn(
            "Relying on GidgetHub's event class interfaces will be deprecated "
            "in the future releases. Please use 'payload' attribute to access "
            'the event name instead.',
            category=PendingDeprecationWarning,
            stacklevel=2,
        )
        return self.payload

    @property
    def event(self):
        """Event name alias."""
        warnings.warn(
            "Relying on GidgetHub's event class interfaces will be deprecated "
            "in the future releases. Please use 'name' attribute to access "
            'the event name instead.',
            category=PendingDeprecationWarning,
            stacklevel=2,
        )
        return self.name


class GidgetHubActionEvent(GidgetHubEventMixin, GitHubEvent):
    """GitHub Action event wrapper exposing GidgetHub attrs."""


class GidgetHubWebhookEvent(GidgetHubEventMixin, GitHubWebhookEvent):
    """GitHub HTTP event wrapper exposing GidgetHub attrs."""
